/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#include <assert.h>
#include "debugdev.h"
#include <string.h>
/*
 * This is a pseudo-module that allows the software running to send messages
 * to the emulator for logging.  This is meant to accept a byte string from
 * the software and output to the screen for diagnostic purposes.
 *
 * Registers:
 *
 * DATA - The output byte stream to be displayed.  It is displayed once a
 *    \000 or \n is sent.
 * MODEB0 - The least-significant byte of the debug mode
 * MODEB1 - The 2nd byte of the debug mode
 * MODEB2 - The 3rd byte of the debug mode
 * MODEB3 - The most-significant byte of the debuge mode
 *
 * Parameters:
 *
 * DEBUGMODE - Either a number (numerical mask) or list of names to assemble
 *    the mask.  See dbg_modes.h in TinyOS for the list.
 */

/* Declare local statics */
static void debug_destroy_device( void *device );
static uint8_t debug_data_read(void *dev);
static void debug_data_write(void *dev, uint8_t value);
static uint8_t debug_byte0_read(void *dev);
static void debug_byte0_write(void *dev, uint8_t value);
static uint8_t debug_byte1_read(void *dev);
static void debug_byte1_write(void *dev, uint8_t value);
static uint8_t debug_byte2_read(void *dev);
static void debug_byte2_write(void *dev, uint8_t value);
static uint8_t debug_byte3_read(void *dev);
static void debug_byte3_write(void *dev, uint8_t value);

/* Quick little function to convert an integer 0<=n<16 into a hex digit
 * '0'-'9','A'-'F' */
static inline char digit2hex(int digit) {
    assert((digit >= 0) && (digit < 16));
    return ((digit < 10) ? (digit + '0') : (digit + 'A' - 10));
}

/* The modes currently assigned for the TinyOS debugging log */
static const char* g_modes[] = {
    "DBG_BOOT",    "DBG_CLOCK", "DBG_TASK",    "DBG_SCHED",
    "DBG_SENSOR",  "DBG_LED",   "DBG_CRYPTO",  "DBG_ROUTE",
    "DBG_AM",      "DBG_CRC",   "DBG_PACKET",  "DBG_ENCODE",
    "DBG_RADIO",   "DBG_LOG",   "DBG_ADC",     "DBG_I2C",
    "DBG_UART",    "DBG_PROG",  "DBG_SOUNDER", "DBG_TIME",
    "DBG_RESERVED","DBG_SIM",   "DBG_QUEUE",   "DBG_SIMRADIO",
    "DBG_HARD",    "DBG_MEM",   "DBG_RESERVED","DBG_USR1",
    "DBG_USR2",    "DBG_USR3",  "DBG_TEMP",    "DBG_ERROR",
    "DBG_RESERVED"
};

/* Given a debugging mode select, determine the index into the above mode
 * listing */
static inline int debug_getmode(uint32_t mode) {
    int pos;
    assert(mode != 0);
    for(pos = 0; !GET_FLAG(mode, 1); pos++, mode >>= 1);
    return pos;
}

/* Use the standard dbg_modes.h selection */
static const struct {
    const char* name;
    uint64_t value;
} g_modelookup[] = {
    DBG_NAMETAB
};


/*
 * IO Manager interface
 */

device_obj* create_device(char *name) {
    debug_device *new_dev;

    /* Create the object for the emulator */
    new_dev = (debug_device*)malloc(sizeof(debug_device));
    if(new_dev == NULL) {
        return NULL;
    }
    memset(new_dev, 0, sizeof(debug_device));

    /* Create the initial display buffer */
    new_dev->buffer = (char*)malloc(64);
    new_dev->buflen = 64;
    new_dev->bufused = 0;
    if(new_dev->buffer == NULL) {
        free(new_dev);
        return NULL;
    }

    /* Initialize */
    new_dev->device.next = NULL;
    new_dev->device.destroy = debug_destroy_device;
    new_dev->device.dev = new_dev;

    new_dev->data.read = debug_data_read;
    new_dev->data.write = debug_data_write;
    new_dev->data.dev = new_dev;

    new_dev->modeb0.read = debug_byte0_read;
    new_dev->modeb0.write = debug_byte0_write;
    new_dev->modeb0.dev = new_dev;
    new_dev->modeb1.read = debug_byte1_read;
    new_dev->modeb1.write = debug_byte1_write;
    new_dev->modeb1.dev = new_dev;
    new_dev->modeb2.read = debug_byte2_read;
    new_dev->modeb2.write = debug_byte2_write;
    new_dev->modeb2.dev = new_dev;
    new_dev->modeb3.read = debug_byte3_read;
    new_dev->modeb3.write = debug_byte3_write;
    new_dev->modeb3.dev = new_dev;

    new_dev->lastbyte = 0;
    new_dev->mode = 0;
    new_dev->mask = 0;

    return &new_dev->device;
}

register_obj *add_register(void *dev, char *name) {
    debug_device *dbg = (debug_device*)dev;

    if(strcasecmp(name, "DATA") == 0) {
        return &(dbg->data);
    } else if(strcasecmp(name, "MODEB0") == 0) {
        return &(dbg->modeb0);
    } else if(strcasecmp(name, "MODEB1") == 0) {
        return &(dbg->modeb1);
    } else if(strcasecmp(name, "MODEB2") == 0) {
        return &(dbg->modeb2);
    } else if(strcasecmp(name, "MODEB3") == 0) {
        return &(dbg->modeb3);
    }

    return NULL;
}

error_t set_param(void *dev, char *key, char *value) {
    debug_device *dbg = (debug_device*)dev;

    if(strcasecmp(key, "DEBUGMODE") == 0) {
        char *next;
        int i;

        next = strtok(value, " ,:;+");
        while(next != NULL) {
            /* Look up this particular term */
            for(i = 0; (g_modelookup[i].name != NULL)
                    && strcasecmp(g_modelookup[i].name, next); i++);
            dbg->mask |= g_modelookup[i].value;
            /* Next in list */
            next = strtok(NULL, " ,:;+");
        }
        return 0;
    }

    return -1;
}

static void debug_destroy_device(void *dev) {
    debug_device *dbg = (debug_device*)dev;

    free(dbg->buffer);
    free(dbg);
}

/*
 * Register I/O
 */

static uint8_t debug_data_read(void *dev) {
    debug_device *dbg = (debug_device*)dev;
    return dbg->lastbyte;
}

static void debug_data_write(void *dev, uint8_t value) {
    debug_device *dbg = (debug_device*)dev;
    static const size_t s_growby = 64;

    if(value == 0) {
        /* Ignore this character */
    } else if(value == '\n') {
        /* Check for special EOL characters */
        if((dbg->bufused > 0) && (dbg->mode)) {
            /* Make sure this is a mode we will display */
            int mode = debug_getmode(dbg->mode);
            if(GET_FLAG(dbg->mask, dbg->mode)) {
                /* We are guaranteed to have space for the trailing 0 */
                dbg->buffer[dbg->bufused] = 0;
                dbg->device.device_event(&(dbg->device), "[%s] %s\n",
                        g_modes[mode], dbg->buffer);
            }
        }
        dbg->bufused = 0;
    } else if(((value < 32) || (value > 126)) && (value != '\t')) {
        /* Check for unprintables */
        if((dbg->bufused + 5) >= dbg->buflen) {
            /* Enlarge buffer */
            void* newbuf = realloc(dbg->buffer, dbg->buflen + s_growby);
            if(newbuf == NULL) {
                /* Hmmm...ran out of memory (but buffer is still allocated) */
                dbg->device.device_debug(&(dbg->device), V_NORMAL,
                        "Out of memory!!!\n");
                return;
            }
            dbg->buffer = (char*)newbuf;
            dbg->buflen += s_growby;
        }
        /* Append a <XX> to the end */
        dbg->buffer[dbg->bufused++] = '<';
        dbg->buffer[dbg->bufused++] = digit2hex(value >> 4);
        dbg->buffer[dbg->bufused++] = digit2hex(value & 0x0F);
        dbg->buffer[dbg->bufused++] = '>';
    } else {
        /* Normal character */
        if((dbg->bufused + 2) >= dbg->buflen) {
            /* Enlarge buffer */
            void *newbuf = realloc(dbg->buffer, dbg->buflen + s_growby);
            if(newbuf == NULL) {
                /* Out of memory */
                dbg->device.device_debug(&(dbg->device), V_NORMAL,
                        "Out of memory!!!\n");
                return;
            }
            dbg->buffer = (char*)newbuf;
            dbg->buflen += s_growby;
        }
        /* Append character */
        dbg->buffer[dbg->bufused++] = value;
    }
}

static uint8_t debug_byte0_read(void *dev) {
    debug_device *dbg = (debug_device*)dev;
    return (dbg->mode & 0x000000FF);
}

static void debug_byte0_write(void *dev, uint8_t value) {
    debug_device *dbg = (debug_device*)dev;
    uint32_t oldmode = dbg->mode;

    dbg->mode = (dbg->mode & 0xFFFFFF00) | value;
    if(dbg->mode != oldmode) {
        dbg->bufused = 0;
    }
}

static uint8_t debug_byte1_read(void *dev) {
    debug_device *dbg = (debug_device*)dev;
    return (dbg->mode & 0x0000FF00) >> 8;
}

static void debug_byte1_write(void *dev, uint8_t value) {
    debug_device *dbg = (debug_device*)dev;
    uint32_t oldmode = dbg->mode;

    dbg->mode = (dbg->mode & 0xFFFF00FF) | ((uint32_t)value << 8);
    if(dbg->mode != oldmode) {
        dbg->bufused = 0;
    }
}

static uint8_t debug_byte2_read(void *dev) {
    debug_device *dbg = (debug_device*)dev;
    return (dbg->mode & 0x00FF0000) >> 16;
}

static void debug_byte2_write(void *dev, uint8_t value) {
    debug_device *dbg = (debug_device*)dev;
    uint32_t oldmode = dbg->mode;

    dbg->mode = (dbg->mode & 0xFF00FFFF) | ((uint32_t)value << 16);
    if(dbg->mode != oldmode) {
        dbg->bufused = 0;
    }
}

static uint8_t debug_byte3_read(void *dev) {
    debug_device *dbg = (debug_device*)dev;
    return (dbg->mode & 0xFF000000) >> 24;
}

static void debug_byte3_write(void *dev, uint8_t value) {
    debug_device *dbg = (debug_device*)dev;
    uint32_t oldmode = dbg->mode;

    dbg->mode = (dbg->mode & 0x00FFFFFF) | ((uint32_t)value << 24);
    if(dbg->mode != oldmode) {
        dbg->bufused = 0;
    }
}
