/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#include <gtk/gtk.h>
#include <inttypes.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "../defs.h"

#include "chip_data.h"
#include "chip_data_struct.h"
#include "table.h"
#include "interface.h" /* for notebook_chips */
#include "atemu.h"
#include "func_data.h"
#include "gtk_helper.h" /* text_buffer_printf */
#include "breakpoints.h"
#include "type_data.h"

#define XATDB_DEFAULT_VAR_SIZE 4

chip_data *chip_array = NULL;
int chip_array_cnt = 0;
int curr_chip = 0;
int cfile_read;

/***************************************/
/* comparison functions for the tables */
/***************************************/

static int addr_compare( const void *a, const void *b )
{
    struct tuple_line_addr *at, *bt;

    at = (struct tuple_line_addr *) a;
    bt = (struct tuple_line_addr *) b;

    return ( at->addr - bt->addr );
}

static int line_compare(const void *a, const void *b){
    struct tuple_line_addr *at, *bt;

    at = (struct tuple_line_addr *) a;
    bt = (struct tuple_line_addr *) b;

    return at->line - bt->line;
}

static int addr_compare_range( const void *a, const void *b )
{
    struct tuple_file_addr *at, *bt;

    at = (struct tuple_file_addr *) a;
    bt = (struct tuple_file_addr *) b;

    if ( bt->addr_begin > at->addr_begin )
        return -1;

    if ( bt->addr_end <= at->addr_begin )
        return 1;

    return 0;
}

static int file_line_compare(const void *a, const void *b){
    struct tuple_file_addr *at, *bt;
    int compare;

    at = (struct tuple_file_addr *) a;
    bt = (struct tuple_file_addr *) b;

    compare = strcmp(at->fname, bt->fname);
    if(compare == 0){
        return at->line - bt->line;
    }
    return compare;
}

static int cfile_compare(const void *a, const void *b){
    struct tuple_cfile_textbuffer *at, *bt;

    at = (struct tuple_cfile_textbuffer *)a;
    bt = (struct tuple_cfile_textbuffer *)b;

    return strcmp(at->fname, bt->fname);
}

static int cfile_index_compare(const void *a, const void *b){
    struct tuple_cfile_textbuffer *at, *bt;

    at = (struct tuple_cfile_textbuffer *)a;
    bt = (struct tuple_cfile_textbuffer *)b;

    return at->menu_index - bt->menu_index;
}

static int global_compare(const void *a, const void *b){
    struct tuple_global_addr *at, *bt;

    at = (struct tuple_global_addr *) a;
    bt = (struct tuple_global_addr *) b;

    return ( at->notify_id - bt->notify_id );
}

static int global_name_compare(const void *a, const void *b){
    struct tuple_global_addr *at, *bt;

    at = (struct tuple_global_addr *) a;
    bt = (struct tuple_global_addr *) b;

    return ( strcmp( at->name, bt->name ) );
}

static int global_id_compare(const void *a, const void *b){
    struct tuple_global_addr *at, *bt;

    at = (struct tuple_global_addr *) a;
    bt = (struct tuple_global_addr *) b;

    return ( at->notify_id - bt->notify_id );
}

/***************************************/
/* destructors for tuples that need it */
/***************************************/

void destroy_file_addr( void * a )
{
    struct tuple_file_addr * tup = (struct tuple_file_addr *) a;

    free( tup->fname );
    free( tup );
}

void destroy_global( void * a )
{
    struct tuple_global_addr * tup = (struct tuple_global_addr *) a;

    free( tup->name );
    free( tup );
}

void destroy_cfile_buff( void * a )
{
    struct tuple_cfile_textbuffer * tup;
    tup = (struct tuple_cfile_textbuffer *) a;

    free( tup->fname );
    g_object_unref(tup->textbuff);
    free( tup );
}

/***********************/
/* adds for the tables */
/***********************/

void add_addr_file_mapping( int chipid, long int begin, long int end,
    long int line, char * file )
{
    struct tuple_file_addr * to_add;

    to_add = (struct tuple_file_addr *) malloc( sizeof(struct tuple_file_addr) );
    if ( to_add == NULL )
    {
        printf( "malloc failed in add_addr_mapping.\n" );
        return;
    }

    to_add->addr_begin = begin;
    to_add->addr_end = end;
    to_add->line = line;
    to_add->fname = strdup( file );

    table_store( chip_array[chipid].addr_to_c_line, to_add );
    table_store( chip_array[chipid].c_line_to_addr, to_add );
}

void add_addr_line_mapping( int chipid, long int addr, long int line )
{
    struct tuple_line_addr * to_add;

    to_add = (struct tuple_line_addr *) malloc( sizeof(struct tuple_line_addr) );
    if ( to_add == NULL )
    {
        printf( "malloc failed in add_addr_mapping.\n" );
        return;
    }

    to_add->addr = addr;
    to_add->line = line;

    table_store( chip_array[chipid].addr_to_text_line, to_add );
    table_store( chip_array[chipid].text_line_to_addr, to_add );
}

void add_func_data( int chipid, struct func_data *fd )
{
    table_store( chip_array[chipid].functions, fd );
}

void add_type_data( int chipid, struct type_data *td )
{
    table_store( chip_array[chipid].types, td );
}

/* helper for make_cfile_menu */
static int cfile_index;
void translate_cfile_menu(void *node, void *data){
    int *chipid = (int *)data;
    GtkWidget *menuitem;
    struct tuple_cfile_textbuffer *cfile;
    cfile = (struct tuple_cfile_textbuffer *)node;

    /* add it to the menu list */
    menuitem = gtk_menu_item_new_with_label(cfile->fname);
    gtk_widget_show(menuitem);
    gtk_menu_shell_append(GTK_MENU_SHELL(chip_array[*chipid].cfile_select_menu), menuitem);
    
    /* it will have this index number in the list */
    cfile->menu_index = cfile_index;
    table_store(chip_array[*chipid].cfile_index_to_textbuffers, cfile);
    
    cfile_index++;
}

/* sets up menu indices and makes a second mapping table */
void make_cfile_menu(int chipid){
    cfile_index = 0;

    /* do an in-order traversal of the filename list to make a sorted menu and index mapping */
    table_foreach(chip_array[chipid].cfile_to_textbuffers, translate_cfile_menu, &chipid);

    /* select first file in list */
    gtk_option_menu_set_history(GTK_OPTION_MENU(chip_array[chipid].cfile_select), 0);
}

void add_cfile(int chipid, char *fname){
    struct tuple_cfile_textbuffer query, *cfile;

    query.fname = fname;

    cfile = (struct tuple_cfile_textbuffer *)table_find(chip_array[chipid].cfile_to_textbuffers, &query);
    if(cfile == NULL){
        FILE *file_in;
        file_in = fopen(fname, "r");
        if(file_in != NULL){
            char buff[1024];

            /* notify the world that we've read at least one cfle */
            cfile_read = 1;

            /* make the buffer, then add */
            cfile = malloc(sizeof(struct tuple_cfile_textbuffer));
            cfile->fname = strdup(fname);
            cfile->textbuff = gtk_text_buffer_new(tagtable);

            /* should maybe be in a callback */
            while(!xatdb_quit && fgets(buff, sizeof(buff), file_in) != NULL){
                text_buffer_printf(cfile->textbuff, "%s", buff);
                    g_main_context_iteration(NULL, FALSE);
            }
            fclose(file_in);

            if(xatdb_quit){
                return;
            }

            table_store(chip_array[chipid].cfile_to_textbuffers, cfile);
        }
    }
}

static void sign_extend( uint8_t *buff, int max, int len )
{
    int sign, i;

    if ( len < max ) {
        sign = (buff[len - 1] & 0x80) >> 7;

        for( i = len; i < max; i++ )
            buff[i] = 0xff * sign;
    }
}


/* print a type nicely.  if not possible, i.e. not enough space in
    provided buffer, return < 0.
 */
static int pretty_print( char *buff, int buff_size, uint8_t *value,
    struct type_data *td, int style )
{
    int size;
    int sign;
    uint8_t guard[8] = { 0 };

    buff[0] = 0;

    if ( value == NULL )
    {
        snprintf( buff, buff_size, "Not enough information." );
        return 0;
    }

    if ( td )
    {
        size = td->byte_length;
        sign = td->sign_type;

        if ( td->fields != NULL )
        {
            /* structure... do something intelligent here. */
            /* call this function for each field. append output */
            /* remmber to do something about nested structures... */
        }
    } else {
        size = XATDB_DEFAULT_VAR_SIZE;
        sign = 0;
    }

    memcpy( guard, value, size );

    if (( sign == 0 ) || ( size > 8 ))
    { 
        char temp[4];
        int curr;
        int b_size = (size * 4) * sizeof(char);

        if ( buff_size < b_size )
            return -1;

        for( curr = 0; curr < size; curr++ )
        {
            if ( ( (curr % 8) == 0 ) && ( curr != 0 ) )
            {
                strncat( buff, "\n", buff_size - strlen(buff) - 1 );
            }

            snprintf( temp, 4, "%02x ", ((uint8_t *)value)[curr] );
            strncat( buff, temp, buff_size - strlen(buff) - 1 );
        }
    } else if ( sign < 0 ) {
        switch( size ) {
            case 8: case 7: case 6: case 5:
                sign_extend( guard, 8, size );
                {
                    int64_t temp = *((int64_t *) &guard[0]);

                    temp = CONVERT_8BYTE(temp);

                    if ( style == 0 )
                    {
                        snprintf( buff, buff_size, "%lld", temp );
                    } 
                    else if (style == 1 ) 
                    {
                        snprintf( buff, buff_size, "0x%016llx", temp );
                    }
                }
                break;

            case 4: case 3:
                sign_extend( guard, 4, size );
                {
                    int32_t temp = *((int32_t *) &guard[0]);

                    temp = CONVERT_4BYTE(temp);

                    if ( style == 0 )
                    {
                        snprintf( buff, buff_size, "%d", temp );
                    } 
                    else if (style == 1 ) 
                    {
                        snprintf( buff, buff_size, "0x%08x", temp );
                    }
                }
                break;

            case 2:
                {
                    int16_t temp = *((int16_t *) &guard[0]);

                    temp = CONVERT_2BYTE(temp);
                    if ( style == 0 )
                    {
                        snprintf( buff, buff_size, "%d", temp );
                    } 
                    else if (style == 1 ) 
                    {
                        snprintf( buff, buff_size, "0x%04x", temp );
                    }
                }
                break;

            case 1:
                {
                    int8_t temp = *((int8_t *) &guard[0]);

                    if ( style == 0 )
                    {
                        snprintf( buff, buff_size, "%d", temp );
                    } 
                    else if (style == 1 ) 
                    {
                        snprintf( buff, buff_size, "0x%02x", temp );
                    }
                }
                break;

            default:
                snprintf( buff, buff_size, "Something is seriously wrong." );
                break;
        }
    } else if ( sign > 0 ) {
        switch( size ) {
            case 8: case 7: case 6: case 5:
                {
                    uint64_t temp = *((uint64_t *) &guard[0]);

                    temp = CONVERT_8BYTE(temp);
                    if ( style == 0 )
                    {
                        snprintf( buff, buff_size, "%lld", temp );
                    } 
                    else if (style == 1 ) 
                    {
                        snprintf( buff, buff_size, "0x%016llx", temp );
                    }
                }
                break;

            case 4: case 3:
                {
                    uint32_t temp = *((uint32_t *) &guard[0]);

                    temp = CONVERT_4BYTE(temp);
                    if ( style == 0 )
                    {
                        snprintf( buff, buff_size, "%d", temp );
                    } 
                    else if (style == 1 ) 
                    {
                        snprintf( buff, buff_size, "0x%08x", temp );
                    }
                }
                break;

            case 2:
                {
                    uint16_t temp = *((uint16_t *) &guard[0]);

                    temp = CONVERT_2BYTE(temp);
                    if ( style == 0 )
                    {
                        snprintf( buff, buff_size, "%d", temp );
                    } 
                    else if (style == 1 ) 
                    {
                        snprintf( buff, buff_size, "0x%04x", temp );
                    }
                }
                break;

            case 1:
                {
                    uint8_t temp = *((uint8_t *) &guard[0]);

                    if ( style == 0 )
                    {
                        snprintf( buff, buff_size, "%d", temp );
                    } 
                    else if (style == 1 ) 
                    {
                        snprintf( buff, buff_size, "0x%02x", temp );
                    }
                }
                break;

            default:
                snprintf( buff, buff_size, "Something is seriously wrong." );
                break;
        }
    }

    return 0;
}

void render_global( int chipid, struct tuple_global_addr *glob )
{
    char *error = "Error";
    char buff[128];

    if ( pretty_print( buff, 128, (uint8_t *)glob->value, glob->td,
        glob->type_style ) < 0 )
    {
        gtk_list_store_set( chip_array[chipid].globals, &glob->iter,
            2, error, -1 );
    } else {
        gtk_list_store_set( chip_array[chipid].globals, &glob->iter,
            2, buff, -1 );
    }

}

void update_global( int chipid, int id, long int addr, uint8_t new_value )
{
    struct tuple_global_addr *to_add, query;

    query.notify_id = id;
    to_add = table_find( chip_array[chipid].global_by_id, &query );
    if ( to_add == NULL ) {
        printf( "Tried update entry for unknown global.[id = %d]\n", id );
        return;
    }

    /* if we have the size info... */
    if ( to_add->value != NULL ) {
        ((uint8_t *)to_add->value)[addr - to_add->addr] = new_value;    
    }


    render_global( chipid, to_add );
}

void size_global( int chipid, char *name, struct type_data *td )
{
    struct tuple_global_addr *to_add, query;
    int i;
    int size;

    query.name = name;
    to_add = table_find( chip_array[chipid].global_by_name, &query );
    if ( to_add == NULL ) {
        printf( "Tried fixing a size entry for unknown global.\n" );
        return;
    }

    to_add->td = td;
    to_add->type_style = 0;

    if ( td )
    {
        size = td->byte_length;
    } else {
        /* if we don't know... 32 bits sounds good. */
        size = XATDB_DEFAULT_VAR_SIZE;
    }

    to_add->value = (uint8_t *) calloc( size, sizeof(uint8_t) );

    if ( to_add->value == NULL )
    {
        printf( "calloc failed in size_global.\n" );
        return;
    }

    for( i = 0; i < size; i++ )
        atemu_add_notify( chipid, to_add->addr + i, to_add->notify_id );

}

void add_global( int chipid, char * symbol, long int addr )
{
    static int id = 0;

    GtkTreeIter iter;
    struct tuple_global_addr *to_add;
    char buff[32];

    to_add = (struct tuple_global_addr *) 
        malloc( sizeof(struct tuple_global_addr) );
    if ( to_add == NULL )
    {
        printf( "malloc failed in add_global.\n" );
        return;        
    }

    snprintf( buff, 32, "%08lx", addr );
    buff[31] = '\0';
    gtk_list_store_append( chip_array[chipid].globals, &iter );

    to_add->name = strdup( symbol );
    to_add->addr = addr;
    to_add->iter = iter;
    to_add->notify_id = id++;
    to_add->value = NULL;

    to_add->td = NULL;
    to_add->type_style = 0;

    table_store( chip_array[chipid].global_addr, to_add );
    table_store( chip_array[chipid].global_by_name, to_add );
    table_store( chip_array[chipid].global_by_id, to_add );

    gtk_list_store_set( chip_array[chipid].globals, &iter, 0, symbol,
        1, buff, 3, to_add->notify_id, -1 );
}

/***************************/
/* fetches from the tables */
/***************************/

struct local *id_to_local_data( int chipid, int id )
{
    struct local *ret = NULL;
    GSList *curr;

    curr = (chip_array[chipid].curr_func)->locals;

    while( curr != NULL )
    {
        if ( ((struct local *) curr->data)->id == id )
        {
            ret = ((struct local *) curr->data);
            break;
        }

        curr = curr->next;
    }

    return ret;
}


struct tuple_global_addr *id_to_global_data( int chipid, int id )
{
    struct tuple_global_addr *ret, query;
 
    query.notify_id = id;
    ret = table_find( chip_array[chipid].global_by_id, &query );

    return ret;
}

struct type_data *id_to_type_data( int chipnum, char *id )
{
    struct type_data search, *found;

    search.id = id;
    
    found = table_find( chip_array[chipnum].types, &search );

    if ( found == NULL )
    {
        return NULL;
    }

    return found;
}

int id_to_type_size( int chipnum, char *id )
{
    struct type_data *found;

    found = id_to_type_data( chipnum, id );

    if ( found == NULL )
    {
        return -1;
    }

    return found->byte_length;
}

int address_to_text_line( int chipnum, long int address )
{
    struct tuple_line_addr search, *found;
    
    search.addr = address;

    found = table_find( chip_array[chipnum].addr_to_text_line, &search );

    if ( found == NULL )
    {
        return -1;  
    }

    return found->line;
}

long text_line_to_address(int chipnum, int line){
    struct tuple_line_addr search, *found;
    
    search.line = line;

    found = table_find( chip_array[chipnum].text_line_to_addr, &search );

    if ( found == NULL )
    {
        return -1;  
    }

    return found->addr;
}

long c_line_to_address( int chipnum, char *cfile, int line )
{
    struct tuple_file_addr search, *found;
    
    search.fname = cfile;
    search.line = line;

    found = table_find( chip_array[chipnum].c_line_to_addr, &search );

    if ( found == NULL )
    {
        return -1;
    }

    return found->addr_begin;
}

long address_to_c_line( int chipnum, long int address, char **cfile )
{
    struct tuple_file_addr search, *found;
    
    search.addr_begin = address;

    found = table_find( chip_array[chipnum].addr_to_c_line, &search );

    if ( found == NULL )
    {
        *cfile = NULL;
        return -1;
    }

    *cfile = found->fname;
    return found->line;
}

char * address_to_function( int chipnum, long int address )
{
    struct func_data search, *found;

    search.addr_begin = address;
    
    found = table_find( chip_array[chipnum].functions, &search );

    if ( found == NULL )
    {
        return NULL;
    }

    return found->name;
}

struct func_data * address_to_func_data( int chipnum, long int address )
{
    struct func_data search, *found;

    search.addr_begin = address;
    
    found = table_find( chip_array[chipnum].functions, &search );

    return found;
}

char *cfile_index_to_cfile(int chipnum, int index){
    struct tuple_cfile_textbuffer query, *result;

    query.menu_index = index;

    result = table_find(chip_array[chipnum].cfile_index_to_textbuffers, &query);

    if(result != NULL){
        return result->fname;
    }else{
        return NULL;
    }
}

GtkTextBuffer *cfile_index_to_textbuffer(int chipnum, int index){
    struct tuple_cfile_textbuffer query, *result;

    query.menu_index = index;

    result = table_find(chip_array[chipnum].cfile_index_to_textbuffers, &query);

    if(result != NULL){
        return result->textbuff;
    }else{
        return NULL;
    }
}

GtkTextBuffer *cfile_to_textbuffer(int chipnum, char *fname){
    struct tuple_cfile_textbuffer query, *result;

    query.fname = fname;

    result = table_find(chip_array[chipnum].cfile_to_textbuffers, &query);

    if(result != NULL){
        return result->textbuff;
    }else{
        return NULL;
    }
}

int cfile_to_cfile_index(int chipnum, char *fname){
    struct tuple_cfile_textbuffer query, *result;

    query.fname = fname;

    result = table_find(chip_array[chipnum].cfile_to_textbuffers, &query);

    if(result != NULL){
        return result->menu_index;
    }else{
        return -1;
    }
}

/**************************/
/* chip struct management */
/**************************/

static void chip_create_data( chip_data * data )
{
    int i;
    char addr[16];
    GtkTreeIter   iter;

    /* REGS:    Register Name: Hex. Value: Unsigned Dec. Value (Signed)*/
    data->regs = gtk_list_store_new( 3, G_TYPE_STRING, G_TYPE_STRING,
        G_TYPE_STRING );

    for( i = 0; i < 32; i++ )
    {
        gtk_list_store_append ( GTK_LIST_STORE(data->regs), &iter );

        snprintf( addr, 16, "R%d", i );
        addr[15] = '\0';
        gtk_list_store_set ( GTK_LIST_STORE(data->regs), &iter,
            0, addr, -1);
    }

    gtk_list_store_append ( GTK_LIST_STORE(data->regs), &iter );
    gtk_list_store_set ( GTK_LIST_STORE(data->regs), &iter,
        0, "X", -1);

    gtk_list_store_append ( GTK_LIST_STORE(data->regs), &iter );
    gtk_list_store_set ( GTK_LIST_STORE(data->regs), &iter,
        0, "Y", -1);

    gtk_list_store_append ( GTK_LIST_STORE(data->regs), &iter );
    gtk_list_store_set ( GTK_LIST_STORE(data->regs), &iter,
        0, "Z", -1);

    gtk_list_store_append ( GTK_LIST_STORE(data->regs), &iter );
    gtk_list_store_set ( GTK_LIST_STORE(data->regs), &iter,
        0, "PC",-1);

    gtk_list_store_append ( GTK_LIST_STORE(data->regs), &iter );
    gtk_list_store_set ( GTK_LIST_STORE(data->regs), &iter,
        0, "SP",-1);

    gtk_list_store_append ( GTK_LIST_STORE(data->regs), &iter );
    gtk_list_store_set ( GTK_LIST_STORE(data->regs), &iter,
        0, "SREG",-1);

    /* STACK:   C file : C line */
    data->stack = gtk_list_store_new( 2, G_TYPE_STRING, G_TYPE_STRING );

    /* LOCALS:  Name: Hex. Value: Unsigned Dec. Value (Signed) */
    data->locals = gtk_list_store_new( 4, G_TYPE_STRING, G_TYPE_STRING,
        G_TYPE_STRING, G_TYPE_INT );

    /* GLOBALS: Name: Hex. Value: Unsigned Dec. Value (Signed) */
    data->globals = gtk_list_store_new( 4, G_TYPE_STRING, G_TYPE_STRING,
        G_TYPE_STRING, G_TYPE_INT );
    
    data->code = gtk_text_buffer_new( tagtable );
    data->events = gtk_text_buffer_new( tagtable );
    data->debug = gtk_text_buffer_new( tagtable );   

    /* BREAK:   Type : Hex. Location : (hidden)address */
    data->breakp = gtk_list_store_new( 3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT );
    
    /* WATCH:   Hex. Location : (hidden)address : (hidden) size in bytes */
    data->watchp = gtk_list_store_new( 3, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT );

    /* line <-> address mappings */    
    data->addr_to_text_line = table_create( addr_compare );
    data->text_line_to_addr = table_create( line_compare );
    data->addr_to_c_line = table_create( addr_compare_range );
    data->c_line_to_addr = table_create( file_line_compare );
    data->cfile_to_textbuffers = table_create( cfile_compare );
    data->cfile_index_to_textbuffers = table_create( cfile_index_compare );

    /* symbol mappings */
    data->functions = table_create( func_data_compare );
    data->global_addr = table_create( global_compare );
    data->global_by_name = table_create( global_name_compare );
    data->global_by_id = table_create( global_id_compare );

    /* type mappings */
    data->types = table_create( type_data_compare_id );

    /* breakpoint lists */
    data->asm_break = table_create(asm_break_comp);
    data->source_break = table_create(source_break_comp);
    data->auto_break = table_create(auto_break_comp);
    auto_breaks_active = 0;

    data->curr_func = NULL;

    data->stack_pointer = 0;
    data->status_register = 0;
}

void destroy_chips(){
    int i;
    
    toolbar_sense_set(FALSE);
    
    for(i=0; i < chip_array_cnt; i++){
        gtk_notebook_remove_page(GTK_NOTEBOOK(notebook_chips), 1);

        /* destroy tables, but only free the shared data once */
        table_destroy(chip_array[i].cfile_to_textbuffers, NULL);
        table_destroy(chip_array[i].cfile_index_to_textbuffers, destroy_cfile_buff);
        
        table_destroy(chip_array[i].addr_to_text_line, NULL);
        table_destroy(chip_array[i].text_line_to_addr, free);
        
        table_destroy(chip_array[i].addr_to_c_line, NULL);
        table_destroy(chip_array[i].c_line_to_addr, destroy_file_addr);
        
        table_destroy( chip_array[i].global_by_name, NULL );
        table_destroy( chip_array[i].global_by_id, NULL );
        table_destroy( chip_array[i].global_addr, destroy_global );

        table_destroy( chip_array[i].functions, free_func_data );

        table_destroy( chip_array[i].types, free_type_data );

        table_destroy(chip_array[i].asm_break, free);
        table_destroy(chip_array[i].source_break, free);
        table_destroy(chip_array[i].auto_break, free);
    }

    chip_array_cnt=0;
    
    free(chip_array);
    chip_array = NULL;

    chips_destroyed = 1;
}
     
void create_all_chips( int ntab )
{    
    int i;

    chip_array = (chip_data *) malloc( ntab * sizeof(chip_data) );
    if ( !chip_array )
    {
        printf("Error! Can't allocate memory for chip data.\n" );
        return;
    }
     
    chip_array_cnt = ntab;
 
    for ( i = 0; i < ntab; i++ )
    {
        chip_create_data( &chip_array[i] );
        create_chip_tab( notebook_chips, i );
        atemu_unassemble(i);
    }
    atemu_load_finish();

    chips_destroyed = 0;
    cfile_read = 0;
}

void set_register(int chipn, int line, long int value, int h_field, int d_field)
{
    char hex[32], dec[32], form[32];
    GtkTreeIter iter;

    if ( ! gtk_tree_model_iter_nth_child ( GTK_TREE_MODEL(chip_array[chipn].regs),
         &iter, NULL, line ) )
    {
        printf( "Couldn't find %d-th entry in the regs table for chip %d.\n",
            line, chipn );
        return;
    }

    snprintf( form, 32, "0x%%0%dx", h_field );
    form[31] = '\0';
    snprintf( hex, 32, form, value );
    hex[31] = '\0';
//    snprintf( form, 32, "%%0%dd", d_field );
//    snprintf( dec, 32, form, value );
    snprintf( dec, 32, "%ld", value );
    dec[31] = '\0';

    gtk_list_store_set ( GTK_LIST_STORE(chip_array[chipn].regs), &iter, 
        1, hex, 2, dec, -1);
}

void set_register_name( int chipn, int line, char *value )
{
    char name[128];
    GtkTreeIter iter;


    if ( gtk_tree_model_iter_nth_child( GTK_TREE_MODEL(chip_array[chipn].regs), 
        &iter, NULL, line ) )
    {
        if ( value != NULL )
        {
            snprintf( name, 128, "%s", value );
            name[127] = '\0';
        } else {
            snprintf( name, 128, "R%d", line );
            name[127] = '\0';
        }

        gtk_list_store_set ( GTK_LIST_STORE(chip_array[chipn].regs),
            &iter, 0, name, -1 );
    }
}

void refresh_registers( int chipn, long data_addr, uint8_t buff[16] )
{
    int i;

    if ( data_addr <= 0x10 ) {
        for( i = data_addr; i < (data_addr + 16); i++ )
        {    
            set_register( chipn, i, buff[i - data_addr], 2, 3 );
        }
    }

    if ( data_addr == 0x10 ) {
        unsigned int big;

        /* X */
        big = (buff[10] & 0xff) | ( (buff[11] & 0xff) << 8 );
        set_register( chipn, REGS_TAB_X, big, 4, 5 );

        /* Y */
        big = (buff[12] & 0xff) | ( (buff[13] & 0xff) << 8 );
        set_register( chipn, REGS_TAB_Y, big, 4, 5 );

        /* Z */
        big = (buff[14] & 0xff) | ( (buff[15] & 0xff) << 8 );
        set_register( chipn, REGS_TAB_Z, big, 4, 5 );
    }

    if ( data_addr == 0x50 ) {
        unsigned int big;

        /* STACK */
        big = (buff[0xd] & 0xff) | ( (buff[0xe] & 0xff) << 8 );
        set_register( chipn, REGS_TAB_STACK, big, 4, 5 );
        chip_array[chipn].stack_pointer = big;

        /* SREG */
        big = buff[0xf] & 0xff;
        set_register( chipn, REGS_TAB_SREG, big, 2, 3 );
        chip_array[chipn].status_register = big;
    }

}

void render_local( int chipid, struct local *loc, GtkTreeIter *iter )
{
    char *error = "Error";
    char buff[128];

    if ( pretty_print( buff, 128, (uint8_t *)loc->value, loc->td,
        loc->type_style ) < 0 )
    {
        gtk_list_store_set( chip_array[chipid].locals, iter, 
            1, error, -1 );
    } else {
        gtk_list_store_set( chip_array[chipid].locals, iter, 
            1, buff, -1 );
    }
}

void new_locals( int chipid, long int addr, uint8_t buff[16] )
{
    GSList *curr;
    GtkTreeIter iter;
    struct local * loc;
    int cnt = 0;
    struct func_data * fd;

    if ( chip_array[chipid].curr_func != NULL )
    {
        fd = chip_array[chipid].curr_func;

        for( curr = fd->locals; curr != NULL; curr = curr->next )
        {
            long int laddr;
            long int num_para;
            long int new_para;
            long int curr_para;
            int size;

            loc = (struct local *) curr->data;

            if ( loc->td != NULL )
            {
                size = loc->td->byte_length;
            } else {
                size = XATDB_DEFAULT_VAR_SIZE;
            }

            laddr = loc->location;

            if ( loc->type == LOCAL_TYPE_STACK )
            {
                laddr += chip_array[chipid].stack_pointer;
            }

            num_para = ( (laddr + size - 1) / 16 ) - ( laddr / 16 );
            curr_para = ( laddr / 16 );
            new_para = ( addr / 16 );

            if ( ( new_para >= curr_para ) && 
                 ( new_para <= curr_para + num_para ) )
            {
                int dest_offset;
                int src_offset;
                int length;

                if ( new_para == curr_para )
                {
                    src_offset = laddr % 16;
                    dest_offset = 0;

                    length = ( size < ( 16 - (laddr % 16) )) ? (size) : 
                        ( 16 - (laddr % 16) );
                } else {
                    src_offset = 0;
                    dest_offset = (16 - (laddr % 16)) + 
                        16 * (new_para - curr_para - 1);

                    length = ( ( size - dest_offset ) < 16 ) ?
                        ( size - dest_offset ) : ( 16 );
                }

                if ( loc->value != NULL )
                {
                    memcpy( loc->value + dest_offset, buff + src_offset,
                        length );
                }

                gtk_tree_model_iter_nth_child(
                    GTK_TREE_MODEL(chip_array[chipid].locals),
                    &iter, NULL, cnt );

                render_local( chipid, loc, &iter );
            }

            cnt++;
        }
        
    }
}

void dump_locals( int chipid, struct func_data *fd )
{
    GSList *curr;
    struct local * loc;
    long int old_para = -1;
    long int curr_para;

    if ( chip_array[chipid].curr_func != NULL )
    {
        for( curr = fd->locals; curr != NULL; curr = curr->next )
        {
            long int laddr;
            int size, para;

            loc = (struct local *) curr->data;

            laddr = loc->location;

            if ( loc->td != NULL )
            {
                size = loc->td->byte_length;
            } else {
                size = XATDB_DEFAULT_VAR_SIZE;
            }

            if ( loc->type == LOCAL_TYPE_STACK )
            {
                laddr += chip_array[chipid].stack_pointer;
            }

            para = ( (laddr + size - 1) / 16 ) - ( laddr / 16 );

            /* the dump dumps blocks of 16 bytes, with 16 byte alignment */
            for( curr_para = 0; curr_para <= para; curr_para++ )
            {
                if((curr_para + (laddr / 16)) != old_para){
                    atemu_dump_sram( chipid, (curr_para + (laddr / 16)) * 16 );
                }
            }

            /* just use the first one... chances are it doesn't cross
                   a boundary.
             */
            old_para = (laddr / 16);
        }
    }
}

void update_locals( int chipid, long int addr )
{
    struct func_data *fd;
    GSList *curr;
    GtkTreeIter iter;
    GtkTextMark *mark;

    fd = address_to_func_data( chipid, addr );

    /* We were in this func the step before... */
    if ( fd == chip_array[chipid].curr_func )
    {
        /* nothing to do! */
        dump_locals( chipid, fd );

        return;
    }

    chip_array[chipid].curr_func = fd;

    /* Throw some stuff to console: remind us what function we're in. */
    if ( fd != NULL )
    {
        mark = text_buffer_printf( textbuff_console, 
            "Entering function: %s\n", fd->name );
    } else {
        mark = text_buffer_printf( textbuff_console,
            "Entering function: Unknown\n");
    }

    gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(textview_console),
        mark, 0.0, FALSE, 0.0, 0.0);

    /* must have entered a function different from prev step */

    /* destroy old "Locals" table */
    gtk_list_store_clear( GTK_LIST_STORE( chip_array[chipid].locals ) );

    /* create new "Locals" table (check for "Unknown") */
    if ( fd != NULL )
    {
        struct local *loc;

        for( curr = fd->locals; curr != NULL; curr = curr->next )
        {
            loc = (struct local *) curr->data;

            gtk_list_store_append ( GTK_LIST_STORE(chip_array[chipid].locals),
                &iter );

            gtk_list_store_set( GTK_LIST_STORE( chip_array[chipid].locals ), 
                &iter, 0, loc->name, 3, loc->id, -1 );
        }

        dump_locals( chipid, fd );
    }
}

void add_stack_item( int chipid, long int addr )
{
    char *file;
    char *func;
    char buff[128];
    long int line;
    GtkTreeIter iter;

    line = address_to_c_line( chipid, addr, &file );
    func = address_to_function( chipid, addr );

    if (( line == -1 ) || ( func == NULL )) {
        snprintf( buff, 128, "%08lx\tUnknown", addr ); 
        buff[127] = '\0';
    } else {
//        snprintf( buff, 128, "%08lx\t%s : %ld", addr, file, line ); 
        snprintf( buff, 128, "%08lx\t%s", addr, func ); 
        buff[127] = '\0';
    }

    gtk_list_store_append( chip_array[chipid].stack, &iter );
    gtk_list_store_set( chip_array[chipid].stack, &iter, 0, buff, -1 );
}
